
/* tiffstuff.c is a misnomer. This file sure contains a lot of TIFF-related functions,
	but over time it grew to accommodate other formats, including raw, vsrt, DX,
	and pnm. 
	So this file contains a large collection of internal file handling functions. */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <tiffio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defs.h"
#include "interface.h"
#include "functions.h"
#include "application.h"
#include "dlgbox.h"
#include "gtkhelpers.h"
#include "tiffstuff.h"

/*********************************************************************/

static const char *photoNames[] = 
{
    "min-is-white",				/* PHOTOMETRIC_MINISWHITE */
    "min-is-black",				/* PHOTOMETRIC_MINISBLACK */
    "RGB color",				/* PHOTOMETRIC_RGB */
    "palette color (RGB from colormap)",	/* PHOTOMETRIC_PALETTE */
    "transparency mask",			/* PHOTOMETRIC_MASK */
    "separated",				/* PHOTOMETRIC_SEPARATED */
    "YCbCr",					/* PHOTOMETRIC_YCBCR */
    "7 (0x7)",
    "CIE L*a*b*",				/* PHOTOMETRIC_CIELAB */
};
#define	NPHOTONAMES	(sizeof (photoNames) / sizeof (photoNames[0]))


char tifflog[16384];	/* To display whatever info we consider useful */
#define ADDLOG(C)	strcat (tifflog, C)

#define tiffwnd_title "Tiff Import Info"

static const char *filetype_names[] = 
{
	"Undefined",
	"TIFF (Intel)",
	"TIFF (Motorola)",
	"VSRT",
	"VFLT",
	"DICOM",
	"PGM",
	"PPM"
};
#define	NFILETYPES	(sizeof (filetype_names) / sizeof (filetype_names[0]))



/*================================================================*/


/* TODO list for the TIFF formats.

   1) Contiguous tiled, separate tiled, and separate strip 
   		can't read 12-bit formats or 1-bit fax formats.
   2) Separate tiled and strip can't really handle multiple samples
   		per pixel - should be fixed
*/   


/* there are 4 different "packaging" modes:

	1. Contiguous tiled data
	2. Separate tiled data
	3. Contiguous strip data
	4. Separate strip data
	
	Tiled data mode divides the image into rectangles (tiles) that need
	to be read one by one and fit into the right spot on the image.
	Strip data contains a full scan line (usually over the full horizontal
	width), so we may read strip by strip and fill the image.
	Contiguous and separate has significance in RGB images only.
	Contiguous data is stored as triplets (RGB-RGB-...) while
	separate data contains a strip (or tile) of RRRR, then of GGGG
	then of BBBB etc.
*/

/****************** TIFF READ CORE **************************/


/* Contiguous tiled data stores data as tiles in rgb-rgb-rgb or 
   rgba-rgba-rgba... format. Tiled is the preferred storage mode for
   3D images, as one tile represents one slice
*/


int TIFFReadContigTileData(TIFF* tif, int elemsize, int bpp, void* dataptr)
{
unsigned char *buf;
tsize_t rowsize; 
long tilesize, tilecnt;
char s[255];
unsigned char* p1;
unsigned char* p2;
uint32 tw, th, td, w, h, d;
uint32 x,y,z;
uint32 row, col, slice;


	tilesize = TIFFTileSize(tif); tilecnt=0;
	rowsize = TIFFTileRowSize(tif);
	buf = (unsigned char *)_TIFFmalloc(tilesize);
	sprintf (s, "Contiguous tiled: Tile size is %li\n", tilesize);
	ADDLOG (s);

	if (buf) 
	{
		d=1; td=1; 	/* Default the values in case those tags are undef'd */
		TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
		TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
		TIFFGetField(tif, TIFFTAG_IMAGEDEPTH, &d);
		TIFFGetField(tif, TIFFTAG_TILEWIDTH, &tw);
		TIFFGetField(tif, TIFFTAG_TILELENGTH, &th);
		TIFFGetField(tif, TIFFTAG_TILEDEPTH, &td);
		
		dp (1, "Getting tile: Image (%li %li %li) Tile (%li %li %li). Allocated mem: %li bytes\n",
							w,h,d, tw,th,td, tilesize);
		
		for (slice=0; slice < d; slice += td)
		{
			for (row = 0; row < h; row += th) 
			{
				for (col = 0; col < w; col += tw) 
				{
					if (TIFFReadTile(tif, buf, col, row, slice, 0) < 0) 
					{
						ADDLOG ("Error reading tiled data");
						return -1;
					} 
					else 
					{
						dp (3, "Tile at row=%d, col=%d, slice=%d\n",row,col,slice);
						p2=buf; tilecnt++;
						/* Copy element by element. Make sure we don't exceed bounds */
						for (z=slice; z<slice+td; z++)
							for (y=row; y<row+th; y++)
								for (x=col; x<col+tw; x++)
								{
									if ((x<w) && (y<h))
									{
										p1 = dataptr + elemsize*(x+w*(y+h*z));
										memcpy (p1, p2, elemsize);
									}
									p2 += elemsize;
								}
					}
				}
			}
		}
		_TIFFfree(buf);
	}
	sprintf (s, "%li tiles read\n", tilecnt);
	ADDLOG (s);
	
	return 0;
}


/* Separate tiled data has tiles in the sequence R-tile - G-tile - B-tile etc.
   In this case it only makes sense if the data type is uchar (for B_RGB)
   or samples per pixel is 1 since we don't have 3-channel arbitrary types.
   But then, of course, separate mode doesn't make sense.
   We don't really check this, so a 3-channel RGB would end up reading R
   then overwriting R with G then with B. 
   If this is a problem, we might need to read the file into a buffer
   and then convert the buffer into RGB to salvage the image (although
   we lose precison). Of course, if that happens, we could as well
   extend the program by a new type.
*/


int TIFFReadSeparateTileData(TIFF* tif, int elemsize, int bpp, void* dataptr)
{
unsigned char *buf, *p1, *p2;
long tilesize, tilecnt;
tsize_t rowsize;
uint32 tw, th, w, h, d, td, x,y,z;
uint32 row, col, slice;
tsample_t spl, samplesperpixel;
char s[255];


	tilesize = TIFFTileSize(tif); tilecnt=0;
	buf = (unsigned char *)_TIFFmalloc(tilesize);
	rowsize = TIFFTileRowSize(tif);
	sprintf (s, "Separate tiled: Tile size is %li\n", tilesize);
	ADDLOG (s);
	p1 = dataptr;
	
	if (buf) 
	{
		d=1; td=1; 	/* Default the values in case those tags are undef'd */
		TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
		TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
		TIFFGetField(tif, TIFFTAG_IMAGEDEPTH, &d);
		TIFFGetField(tif, TIFFTAG_TILEWIDTH, &tw);
		TIFFGetField(tif, TIFFTAG_TILELENGTH, &th);
		TIFFGetField(tif, TIFFTAG_TILEDEPTH, &td);
		TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);

		dp (1,"TIFF tiled separate: image size (%d,%d,%d), tile size (%d,%d,%d), samples per pixel %d\n",
			w,h,d,tw,th,td,samplesperpixel);

		for (slice=0; slice < d; slice += td)
		{
			for (row = 0; row < h; row += th) 
			{
				for (col = 0; col < w; col += tw) 
				{
					for (spl = 0; spl < samplesperpixel; spl++) 
					{
						if (TIFFReadTile(tif, buf, col, row, slice, spl) < 0) 
						{
							ADDLOG ("Error reading tiled data");
							return -1;
						} 
						else 
						{
							dp (3, "Tile at s=%d, row=%d, col=%d, slice=%d\n",spl,row,col,slice);
							p2=buf; tilecnt++;
							/* Copy element by element. Make sure we don't exceed bounds */
							for (z=slice; z<slice+td; z++)
								for (y=row; y<row+th; y++)
									for (x=col; x<col+tw; x++)
									{
										if ((x<w) && (y<h))
										{
											p1 = dataptr + elemsize*(x+w*(y+h*z));
											memcpy (p1, p2, elemsize);
										}
										p2 += elemsize;
									}

						}

					}
				}
			}
		}
		_TIFFfree(buf);
	}
	sprintf (s, "%li tiles read\n", tilecnt);
	ADDLOG (s);
	
	return 0;
}


/*
	Contiguous strip data stores the strips as sequences of RGB-RGB-RGB...
	or RGBA-RGBA-RGBA... 
	One strip consists of a number (nrow) of scan lines, where the scan line
	is the smallest unit. We have a function that reads one strip of data.
 
*/

int TIFFReadContigStripData(TIFF* tif, int elemsize, int bpp, void* dataptr)
{
unsigned char *buf;
tsize_t scanlinesize = TIFFScanlineSize(tif);
long tiffstripsize;
unsigned char* p;
unsigned short* ps;
char s[255];
int i,j;
uint32 row, h, w, ofs;
uint32 rowsperstrip = (uint32)-1;

	p = dataptr; ps=dataptr;

	tiffstripsize = TIFFStripSize(tif); /* usually scan line (=row) size * rows per strip */
	buf = (unsigned char *)_TIFFmalloc(tiffstripsize);
	if (buf) 
	{

		TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
		TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
		TIFFGetField(tif, TIFFTAG_ROWSPERSTRIP, &rowsperstrip);

		sprintf (s, "Image length %d, image width (height) %d\n",(int)h,(int)w);
		ADDLOG(s);
		sprintf (s, "Scan line size: %d, strip size %li, rows per strip %li\n",
					(int)scanlinesize,tiffstripsize,rowsperstrip);
		ADDLOG (s);
		if (bpp==12)
		{
			sprintf (s, "Performing expansion from 12 to 16 bits/sample\n");
			ADDLOG(s);
		}
		else if (bpp==1)	/* 1 bit per pixel (fax) - expand */
		{
			sprintf (s, "Performing expansion from 1 to 8 bits/sample\n");
			ADDLOG(s);
		}
		else if ((bpp==16)  && (elemsize==6))
		{
			sprintf (s, "Performing conversion from 48-bit RGB to Z stack\n");
			ADDLOG(s);
		}
		

		ofs=0;
		for (row = 0; row < h; row += rowsperstrip) 
		{
			/* nrow is the number of rows to be read. At the end of the image we have less of them */
			uint32 nrows = (row+rowsperstrip > h ? h-row : rowsperstrip);

			/* Read the next strip, decode compression if necessary */
			tstrip_t strip = TIFFComputeStrip(tif, row, 0);
			if (TIFFReadEncodedStrip(tif, strip, buf, nrows*scanlinesize) < 0) 
			{
				ADDLOG ("Error reading strip data");
				return -1;
			} 
		
			if (bpp==12)		/* 12 bit per pixel, must expand */
			{
				ofs=0;
				for (i=0; i<nrows*scanlinesize; i+=3)	/* one element is 3 bytes, thus the increment */
				{
					if (ofs < h*w-1)
					{
					
						/* 3 bytes in buffer (xx-yy-zz) become 2 shorts (0xxy-0yzz) */
					
						dp (3, "Src row %5d  Dest offset %5d   line %5d   %2X%2X%2X to %4X-%4X\n",
								row, ofs, i,
								buf[i],buf[i+1],buf[i+2],
								((unsigned short)buf[i] << 4) + (buf[i+1] >> 4),
								(((unsigned short)buf[i+1] << 8) & 0x0f00) + buf[i+2] );
					
						ps[w*row+ofs  ] = ((unsigned short)buf[i] << 4) + (buf[i+1] >> 4);
						ps[w*row+ofs+1] = (((unsigned short)buf[i+1] << 8) & 0x0f00) + buf[i+2] ;
						ofs += 2;
					}
				}
			}
			else if (bpp==1)	/* 1 bit per pixel (fax) - expand */
			{
				ofs=0;
				for (i=0; i<nrows*scanlinesize; i++)
				{
					if (ofs < h*w-1)
					{
					
						/* expand each bit of buf into one byte of p. */

						for (j=0; j<8; j++)
							p[w*row+ofs+j  ] = (buf[i] >> (7-j)) & 0x01;	/* Big-endian version? */
						ofs += 8;
					}
				}
			
			}
			else if ((bpp=16) && (elemsize==6))	/* 48-bit color RGB */
			{
				ofs=0;
				for (i=0; i<nrows*scanlinesize; i+=6)	/* one element is 3 shorts, thus the increment */
				{
					if (ofs < h*w-1)
					{
					
						/* 3 shorts in buffer are (rrr-ggg-bbb) */

						ps[w*row+ofs      ] = ((unsigned short)(buf[i+1]<<8) + buf[i+0]);
						ps[w*row+ofs+  h*w] = ((unsigned short)(buf[i+3]<<8) + buf[i+2]);
						ps[w*row+ofs+2*h*w] = ((unsigned short)(buf[i+5]<<8) + buf[i+4]);
						ofs += 1;
					}
				}
			
			}
			else			/* No expansion, copy verbatim */
			{
				memcpy (p+elemsize*row*w, buf, nrows*scanlinesize);
			}
			
		}
		_TIFFfree(buf);
	}
	else
		return -1;
		
	return 0;
}

/*
	Separate strip storage format stores RGB and RGBA data in consecutive
	sections of the image. E.g. there is a strip with RRRRRR first,
	then a srtip with GGGGGGGGG and finally with BBBBBBBBBB.
	Like in the tiled separate reader function, this is NOT properly
	handled here and may fail even with RGB 24-bit images (TODO).
	
*/


int TIFFReadSeparateStripData(TIFF* tif, int elemsize, int bpp, void* dataptr)
{
unsigned char *buf;
tsize_t scanline = TIFFScanlineSize(tif);
long tiffstripsize;
char* p;

	p = dataptr;

	tiffstripsize = TIFFStripSize(tif);
	buf = (unsigned char *)_TIFFmalloc(tiffstripsize);
	if (buf) 
	{
		uint32 row, h, w, i;
		uint32 rowsperstrip = (uint32)-1;
		tsample_t s, samplesperpixel;

		TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
		TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
		TIFFGetField(tif, TIFFTAG_ROWSPERSTRIP, &rowsperstrip);
		TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);
		for (row = 0; row < h; row += rowsperstrip) 
		{
			for (s = 0; s < samplesperpixel; s++) 
			{
				uint32 nrow = (row+rowsperstrip > h ?
				    h-row : rowsperstrip);
				tstrip_t strip = TIFFComputeStrip(tif, row, s);
				if (TIFFReadEncodedStrip(tif, strip, buf, nrow*scanline) < 0) 
				{
					ADDLOG ("Error reading strip data");
					return -1;
				} 
				else
				{
					for (i=0; i<nrow*scanline; i++)
					{
						p[s+elemsize*(row*w+i)] = buf[i];
					}
				}
			}
		}
		_TIFFfree(buf);
	}
	else
		return -1;
	
	return 0;
}



/*******************************************

	Generic TIFF write and auxiliary stuff
	
**********************************************/


int getcomp (int comp)	/* Convert internal compression number to official one */
{
	switch (comp)
	{
		case 0:	/* Don't compress */
				return COMPRESSION_NONE;
				break;
		case 1:	/* LZW */
				return COMPRESSION_LZW;
				break;
		case 2:	/* Packbits*/
				return COMPRESSION_PACKBITS;
				break;
		case 3:	/* CCITT Fax v.3 */
				return COMPRESSION_CCITTFAX3;
				break;
		case 4:	/* CCITT Fax v.4 */
				return COMPRESSION_CCITTFAX4;
				break;
		case 5:	/* JPEG */
				return COMPRESSION_JPEG;
				break;
		case 6:	/* Dumb CCITT RLE */
				return COMPRESSION_CCITTRLE;
				break;
		default:
				return COMPRESSION_NONE;
		
	}
}


/* Here, the image data is written to the (opened and initialized)
   TIFF file. We need to distinguish between striped images for 2D
   and tiled images for 3D.
*/

int tiffwrite_core (TIFF *tif, image_type* img, int st, int en)
{
int e,x,y,z;
uint32 tw,th;
char* p; char* destp; char* srcp;
char* tile;

	e = DATALEN(img->imgtype);
	p = img->data + st*e*img->xmax*img->ymax;

	if ((1+en-st) < 2)			/* 2D image */
	{
		TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, img->ymax);
	   	TIFFWriteEncodedStrip(tif, 0, p, e*img->xmax*img->ymax);
	}
	else						/* 3D image, tiled data */
	{
		
		/* The tile size is approximately the size of one slice. We use
		the "official" recommendation for tile size, which may exceed
		image size by up to 15 bytes. In that case, we need to
		prepare a separate buffer where the tile is padded */
		
		tw = img->xmax; th = img->ymax;
		TIFFDefaultTileSize (tif, &tw, &th);	/* get recommended tile size */
		TIFFSetField(tif, TIFFTAG_TILEWIDTH, tw);
		TIFFSetField(tif, TIFFTAG_TILELENGTH, th);
		
		/* For increased efficiency, we trap the (common) case that image
		and tile size are identical */
		
		if ((tw==img->xmax) && (th==img->ymax))
		{
			for (z=st; z<=en; z++)
			{
		    	/* The z parameter is the tile number - each slice is its own unique tile */
				
				TIFFWriteEncodedTile(tif, z-st, p, e*tw*th);
				p += e*img->xmax*img->ymax;
			}
		}
		else
		{
		
			tile = (char*) calloc (tw*th*e,1);
			if (!tile) fatal_memerr();
			
			for (z=st; z<=en; z++)
			{
				/* Copy elements from the original image. The rest is 0 (calloc) */
				
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						destp = tile + e*(x + y*tw);	/* TODO: optimize */
						srcp  = p    + e*(x + y*img->xmax);
						memcpy (destp, srcp, e);
					}


				TIFFWriteEncodedTile(tif, z-st, tile, e*tw*th);
				p += e*img->xmax*img->ymax; /* point to next slice in image */
			}
			free (tile);
		}
	
	}
	return 0;
}


int tiffwrite (char* fname, image_type* img, int comp, char* comment, int st, int en)
{
TIFF *tif;

	tif = TIFFOpen(fname, "w");
	if (!tif) return -1;

	/* Let's start with some general tags */

	TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, img->xmax);
	TIFFSetField(tif, TIFFTAG_IMAGELENGTH, img->ymax);
	if (en-st>0) TIFFSetField(tif, TIFFTAG_IMAGEDEPTH, 1+en-st);
	TIFFSetField(tif, TIFFTAG_COMPRESSION, getcomp(comp));

	if (comment && strlen(comment) > (size_t) 0) 
		TIFFSetField(tif, TIFFTAG_IMAGEDESCRIPTION, comment);

	if (comp == COMPRESSION_CCITTFAX3)
		TIFFSetField(tif, TIFFTAG_GROUP3OPTIONS,
			GROUP3OPT_2DENCODING+GROUP3OPT_FILLBITS);

	TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
	TIFFSetField(tif, TIFFTAG_RESOLUTIONUNIT, (int)2);
	TIFFSetField(tif, TIFFTAG_XRESOLUTION, tiffsave_dpi);
	TIFFSetField(tif, TIFFTAG_YRESOLUTION, tiffsave_dpi);
	TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	
	/* now write format-specific tags and the image data */

	if (img->imgtype==B_CHAR)
	{
		TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1);
    	TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE,  8);
    	TIFFSetField(tif, TIFFTAG_PHOTOMETRIC,    PHOTOMETRIC_MINISBLACK);
		tiffwrite_core (tif, img, st, en);
	}
	else if (img->imgtype==B_SHORT)
	{
		TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1);
    	TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE,  16);
    	TIFFSetField(tif, TIFFTAG_PHOTOMETRIC,    PHOTOMETRIC_MINISBLACK);
		TIFFSetField(tif, TIFFTAG_SAMPLEFORMAT,   SAMPLEFORMAT_INT); /* Signed! */
		tiffwrite_core (tif, img, st, en);
	}
	else if (img->imgtype==B_FLOAT)
	{

		TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1);
    	TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE,  32);
    	TIFFSetField(tif, TIFFTAG_PHOTOMETRIC,    PHOTOMETRIC_MINISBLACK);
		TIFFSetField(tif, TIFFTAG_SAMPLEFORMAT,   SAMPLEFORMAT_IEEEFP);
		tiffwrite_core (tif, img, st, en);
	}
	else if (img->imgtype==B_RGB) 
	{
		TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 3);
		TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE,  8);
		TIFFSetField(tif, TIFFTAG_PHOTOMETRIC,    PHOTOMETRIC_RGB);
		tiffwrite_core (tif, img, st, en);
	}
 
	TIFFClose(tif);

	return 0;
}





/****************************************************************

	is_tiff opens the file and tries to read the magic number.
	Returns 0 if no magic number found, 1 ot 2 if TIFF magic
	
****************************************************************/


char is_tiff (const char* filename)
{
FILE* FP;
unsigned short hdr;

	FP = fopen (filename, "rb");
	if (!FP) return 0;
	
	fread (&hdr, sizeof (hdr), 1, FP);
	fclose (FP);
	
	if (hdr==0x4949) return 1; 	/* Little-endian, (II for Intel) */
	else if (hdr==0x4d4d) return 2;	/* Big-endian (MM for Motorola) */
	else return 0;
	
}

	
int tiffread (image_type* img, const char* filename)
{

TIFF* tiff_handle;
uint32 xsize, ysize, zsize;
int rderr, e, endian;
uint16 bitspersample, samplesperpixel, planarconfig, photometric, sampleformat;
char s[256];
double* pd; float* pf; long idx;

	/* Open the TIFF header again, determine endian-ness.
		Then get a handle using TIFFDOpen */

	endian = is_tiff (filename);		/* 1 = II = Intel little-endian, 2 = MM = Motorola big-endian */
	tifflog[0]=0;
	sprintf (s, "Opening TIFF file %s\n", filename); ADDLOG (s);
	tiff_handle = TIFFOpen (filename, "r");
	if (tiff_handle==0)		 	/* open failed */
	{
		perror ("TIFF open failed\n");
		ADDLOG ("TIFF open failed\n");
		ShowLongMessage (tifflog, tiffwnd_title);
		dp (1, "TIFF open failed\n");
		return -1;
	}
	
	rderr=0;
	
	/* Get some information about the image */
	
	zsize=1;	/* this tag may not be set */
	bitspersample=8; samplesperpixel=1;		/* Default some more */
	TIFFGetField (tiff_handle, TIFFTAG_IMAGEWIDTH, &xsize);
	TIFFGetField (tiff_handle, TIFFTAG_IMAGELENGTH, &ysize);
	TIFFGetField (tiff_handle, TIFFTAG_IMAGEDEPTH, &zsize);
	TIFFGetField (tiff_handle, TIFFTAG_BITSPERSAMPLE, &bitspersample);
	TIFFGetField (tiff_handle, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);
	TIFFGetField (tiff_handle, TIFFTAG_PHOTOMETRIC, &photometric);
	TIFFGetField (tiff_handle, TIFFTAG_SAMPLEFORMAT, &sampleformat);
	
	
	sprintf (s, "Image size (%d x %d x %d). Bits/sample %d, samples/pixel %d\n", 
				(int)xsize, (int)ysize, (int)zsize, bitspersample, samplesperpixel);
	ADDLOG (s);
	sprintf (s, "Sample format (%d) is ",sampleformat);
	ADDLOG(s);
	
	switch (sampleformat)
	{
		case SAMPLEFORMAT_VOID:
			sprintf(s, "void\n");
			break;
		case SAMPLEFORMAT_INT:
			sprintf (s, "signed integer\n");
			break;
		case SAMPLEFORMAT_UINT:
			sprintf(s, "unsigned integer\n");
			break;
		case SAMPLEFORMAT_IEEEFP:
			sprintf(s, "IEEE floating point\n");
			break;
		default:
			sprintf (s, "unknown\n");
	}
	ADDLOG (s);
	
	sprintf (s, "Photometric (%d) is ", photometric); ADDLOG(s);
	if (photometric < NPHOTONAMES)
		sprintf(s, "%s\n", photoNames[photometric]);
	else 
		sprintf (s, "unknown\n");
	ADDLOG(s);
	
	if (endian==1) sprintf (s, "Bit/byte order is Intel little-endian\n");
	else if (endian==2) sprintf (s, "Bit/byte order is Motorola big-endian\n");
	else sprintf (s, "Bit/byte order is %d (unknown) - this is an error and should not happen\n",endian);
	ADDLOG(s);

	/* With this information, define the image parameters */
	
	img->xmax = xsize;
	img->ymax = ysize;
	img->zmax = zsize; if (zsize==0) { zsize=1; img->zmax=1; }
	e=1;
	
	if (samplesperpixel==3) /* RGB, either 8 or 16 bits/pixel */
	{
		if (photometric!=PHOTOMETRIC_RGB)
		{
			sprintf (s, "TIFF WARNING: 3 samples/pixel, but not photometric RGB\n");
			ADDLOG(s);
		}
		
		if (bitspersample==16)	/* 48-bit RGB becomes a 3-slice stack of B_SHORT */
		{
			sprintf (s, "TIFF WARNING: Photometric RGB with 16 bits/pixel.\nCreating image stack.\n");
			ADDLOG(s);
			img->imgtype = B_SHORT;
			e=6;
			img->zmax *=3;
		}
		else		/* Standard 24-bit RGB is handled natively */
		{
			img->imgtype = B_RGB;
			e=3;
		}
	}
	else if (samplesperpixel==1)
	{
		if ((bitspersample==8) || (bitspersample==1))
		{
			img->imgtype = B_CHAR; e=1;
		}
		else if ((bitspersample==16) || (bitspersample==12))
		{
			img->imgtype = B_SHORT; e=2;
		}
		else if (bitspersample==32)
		{
			img->imgtype = B_FLOAT; e=4;
			if (sampleformat!=SAMPLEFORMAT_IEEEFP)
			{
				sprintf (s, "TIFF WARNING: 32 bits/pixel, but not IEEE float\n");
				ADDLOG(s);
			}
		}
		else if (bitspersample==64)
		{
			img->imgtype = B_FLOAT; e=8;
			if (sampleformat!=SAMPLEFORMAT_IEEEFP)
			{
				sprintf (s, "TIFF WARNING: 64 bits/pixel, but not IEEE float\n");
				ADDLOG(s);
			}
			else
				ADDLOG ("TIFF WARNING: Double format file truncated to float precision\n");
		}
		else
		{
			sprintf (s, "Data format (%d bits per sample) not supported.\n",bitspersample);
			ADDLOG (s);
			ShowLongMessage (tifflog, tiffwnd_title);
			dp (0, s);
			return -1;
		}
	}
	else
	{
		sprintf (s, "Data format (%d samples per pixel) not supported.\n",samplesperpixel);
		ADDLOG (s);
		ShowLongMessage (tifflog, tiffwnd_title);
		dp (0, s);
		return -1;
	}

	/* Now allocate image memory */

	pd=NULL;		/* silence a compiler warning */
	if (e==8)		/* Special case. Abuse img->data to temporarily hold the double* */
	{
		pd = calloc (img->xmax*img->ymax*img->zmax,e);
		img->data = pd;
	}
	else
		img->data = calloc (img->xmax*img->ymax*img->zmax,e);  /* Note: allocates 3x too much mem in 48-bit RGB mode */


	if (!img->data) 
	{
		ADDLOG ("FATAL: Insufficient memory\n");
		ShowLongMessage (tifflog, tiffwnd_title);
		fatal_memerr();
	}
	
	TIFFGetField(tiff_handle, TIFFTAG_PLANARCONFIG, &planarconfig);
	if (TIFFIsTiled(tiff_handle)) 
	{
		if (planarconfig == PLANARCONFIG_CONTIG)
		{
			ADDLOG ("Reading TIFF: Tiled data, contiguous configuration\n");
			rderr=TIFFReadContigTileData(tiff_handle,e,bitspersample,img->data);
		}
		else
		{
			ADDLOG("Reading TIFF: Tiled data, non-contiguous configuration\n");
			rderr=TIFFReadSeparateTileData(tiff_handle,e,bitspersample,img->data);
		}
	} 
	else 
	{
		if (planarconfig == PLANARCONFIG_CONTIG)
		{
			ADDLOG ("Reading TIFF: Strip data, contiguous configuration\n");
			rderr=TIFFReadContigStripData(tiff_handle,e,bitspersample,img->data);
		}
		else
		{
			ADDLOG ("Reading TIFF: Strip data, non-contiguous configuration\n");
			rderr=TIFFReadSeparateStripData(tiff_handle,e,bitspersample,img->data);
		}
	}

	/* Handle 64-bit float images. We still have pd set. */
	
	if (e==8)
	{
		img->data = calloc (img->xmax*img->ymax*img->zmax,4);
		pf = img->data; if (!pf) fatal_memerr();
		for (idx=0; idx<img->xmax*img->ymax*img->zmax; idx++)
			pf[idx] = (float) pd[idx];
		free (pd);
	}


	/* When done, determine min and max */
	
	if (!rderr)
		make_minmax (img);

	TIFFClose (tiff_handle);
	ADDLOG ("Tiff file closed\n");
	if (tiffreport) ShowLongMessage (tifflog, tiffwnd_title);

	return rderr;
}




/************************************************************************************

	Strictly, this does not belong into TIFFSTUFF, but it thematically fits -
	VSRT is a neat 16-bit grayscale format, and the functions to read vsrt
	are similar to the functions used for TIFF... so there we are.
	New: Added a new format, vflt which is similar to vsrt, but accepts
	float data (4 bits per pixel)

*************************************************************************************/


/****************************************************************

	is_vsrt opens the file and tries to read the magic number.
	Returns 0 if no magic number found, 1 if vsrt magic
	
****************************************************************/


char is_vsrt (const char* filename)
{
FILE* FP;
char hdr[4];

	FP = fopen (filename, "rb");
	if (!FP) return 0;
	
	fread (&hdr, sizeof (hdr), 1, FP);
	fclose (FP);
	
	if (!strncasecmp (hdr, "VSRT", 4)) return 1;
	if (!strncasecmp (hdr, "VFLT", 4)) return 2;

	return 0;
	
}


/* This function is needed to swap MSB/LSB of a short int */

short swap_msb_lsb (short x)
{
union
{
	unsigned char b[2];
	unsigned short i;
}
u;

unsigned char b;

	u.i=x;
	b = u.b[0]; u.b[0]=u.b[1]; u.b[1]=b;
	return u.i;
}
	


/***************************************************************

	vsrtread obviously reads a vsrt/vflt image file
	
****************************************************************/


int vsrtread (image_type* img, const char* filename, int isflt)
{

FILE* FP;
char buf[256];
int width, height, numslices;
int pt_x, pt_y, pt_z;
int maxval; float fmaxval;
long numelems,i;
int align,e;
short* p;
int min,max,bitmask;


	FP = fopen (filename, "rb");
	if (!FP) return -1;

	fgets (buf, sizeof (buf), FP);	/* read one text line, i.e. until 0x0A */
	fgets (buf, sizeof (buf), FP);	/* read next text line with img size info */
	
	sscanf (buf, "%d %d %d", &width, &height, &numslices);
	dp (1, "VSRT: Dimensions %d x %d x %d \n", width, height, numslices);
	
	fgets (buf, sizeof (buf), FP);	/* read next text line with special point (unused) */
	sscanf (buf, "%d %d %d", &pt_x, &pt_y, &pt_z);
	dp (1, "VSRT: Special point %d x %d x %d \n", pt_x,pt_y,pt_z);
	
	fgets (buf, sizeof(buf), FP);     /* read 4th line of header with maxgray info */
	
	if (isflt)
	{
		sscanf (buf, "%8f", &fmaxval);	/* get maximum gray value (unused)   */
		dp (1, "VFLT: Max gray value %f\n", fmaxval);
	}
	else
	{
		sscanf (buf, "%8d", &maxval);	/* get maximum gray value (unused)   */
		dp (1, "VSRT: Max gray value %d\n", maxval);
	}
	
	fgets(buf, sizeof(buf), FP);    	/* skip final blank line before data begins */
	
	/* Allocate image memory */
	
	img->xmax = width;
	img->ymax = height; 
	img->zmax = numslices;
	if (isflt)
	{
		img->maxval.fmax=fmaxval;		/* tentative */
		img->minval.fmin=0.0;
		img->imgtype=B_FLOAT;
	}
	else
	{
		img->maxval.imax=maxval;		/* tentative */
		img->minval.imin=0;
		img->imgtype=B_SHORT;
	}
	
	if (isflt) e=sizeof (float); else e = sizeof (short);
	
	img->data = calloc (img->xmax*img->ymax*img->zmax,e);
	if (!img->data) 
	{
		fclose (FP);
		fatal_memerr();
	}
	
	/* Before reading data, we align it to an 8-boundary (SGI-legacy) */

	numelems = ftell (FP);
	if (numelems < 0)
	{                                             /* not found? then error */
		dp (1, "VSRT: open: can't get header position\n");
		fclose (FP);
		return -1;
	}
	
	if (numelems % 8)			/* If this is not an 8-boundary... */
	{
		align = 9 - (numelems %8); 	/* Number of bytes to skip */
		fgets (buf, align, FP);		/* dummy read to skip bytes */
	}
	
	numelems = ftell (FP);	
	dp (1, "VSRT: Data starts at position %li (0x%0lX)\n", numelems, numelems);
	
	/* Read the entire data chunk from FP */
	
	if (isflt)
	{
		numelems = fread (img->data, sizeof (float), img->xmax*img->ymax*img->zmax, FP);
		if (numelems != (img->xmax*img->ymax*img->zmax))
		{
			dp (1, "VFLT warning: File size does not match header\n");
		}

		fclose (FP);
		make_minmax (img);
	}
	else
	{
		numelems = fread (img->data, sizeof (short), img->xmax*img->ymax*img->zmax, FP);
		if (numelems != (img->xmax*img->ymax*img->zmax))
		{
			dp (1, "VSRT warning: File size does not match header\n");
		}

		/* Tidy up. We do not use make_minmax here as we need to swap MSB/LSB anyway */

		fclose (FP);

		p = img->data;
		min = max = swap_msb_lsb(*p);
		bitmask=0;
		for (i=0; i < img->xmax*img->ymax*img->zmax; i++)
		{
			*p = swap_msb_lsb (*p);
			if ( (*p) < min) min = *p;
			if ( (*p) > max) max = *p;
			bitmask |= *p;
			p++;
		}
		img->minval.imin=min;
		img->maxval.imax=max;
		dp (1, "VSRT: Found min and max: %d ... %d; Bitmask = %08X\n", min,max, bitmask);
	}
	
	return 0;
	
}

int vsrtwrite (image_type* img, const char* filename, int st, int en)
{

FILE* FP;
char buf[256];
int pt_x, pt_y, pt_z;
long numelems,i;
int align;
short* p1; short* p2; float* p3;
int x,y,z,errcode;

	errcode=0;
	FP = fopen (filename, "wb");
	if (!FP) return -1;

	if (img->imgtype==B_FLOAT)
		sprintf (buf, "vflt\n%li %li %li\n", img->xmax, img->ymax, 1L+en-st);
	else
		sprintf (buf, "vsrt\n%li %li %li\n", img->xmax, img->ymax, 1L+en-st);
	fputs (buf, FP);

	pt_x = pt_y = pt_z = 0;
	sprintf (buf, "%d %d %d\n", pt_x,pt_y,pt_z);
	fputs (buf, FP);
	
	if (img->imgtype==B_FLOAT)
		sprintf (buf, "%8.2f\n", img->maxval.fmax);
	else
		sprintf (buf, "%8d\n", img->maxval.imax);
	fputs (buf, FP);
	
	strcpy (buf, "                                                            \n");
	fputs (buf, FP);
	
	/* Before writing data, we align it to an 8-boundary (SGI-legacy) */

	numelems = ftell (FP);
	dp (1, "VSRTwrite: Unaligned header %li\n", numelems);
	if (numelems < 0)
	{                                             /* not found? then error */
		dp (1, "VSRT: open: can't get header position\n");
		fclose (FP);
		return -1;
	}
	align = 8 - (numelems & 0x07); 	/* Number of bytes to skip */
	for (i=0; i<align; i++) fputc (0x20, FP);
	numelems = ftell (FP);
	dp (1, "VSRTwrite: Aligned header %li\n", numelems);

	if (img->imgtype==B_FLOAT)
	{
		p3 = img->data;
		for (z=st; z<=en; z++)
		{
			numelems = fwrite (&p3[z*img->xmax*img->ymax], sizeof (float), img->xmax*img->ymax, FP);
			if (numelems != (img->xmax*img->ymax))
			{
				dp (1, "VFLT warning: Wrote incorrect number of bytes\n");
				errcode=-1;
			}
		}
	
	}
	else
	{
		/* write data slice by slice, changing from i86-little-endian to SGI-big-endian
		   (which is a SGI legacy) */

		p1 = (short*) calloc (img->xmax * img->ymax, sizeof (short));
		if (!p1) fatal_memerr();
		p2 = img->data;

		for (z=st; z<=en; z++)
		{
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
					p1[x + img->xmax*y] = swap_msb_lsb (p2[x + img->xmax*(y + img->ymax*z)]);

			numelems = fwrite (p1, sizeof (short), img->xmax*img->ymax, FP);
			if (numelems != (img->xmax*img->ymax))
			{
				dp (1, "VSRT warning: Wrote incorrect number of bytes\n");
				errcode=-1;
			}
		}
		free (p1);	
	}


	fclose (FP);

	return errcode;
	
}


/************************************************************************************

	Since we are already talking cruddy formats, this one imports
	vim files from the AstroPhysics X-ray line scanner.
	It starts with a 264 byte header where the last 12 bytes are relevant.
	The vim format is an byte-interlaced format for low and high energy
	respectively.

*************************************************************************************/



int  wim_read (image_type* img, const char* filename)
{

FILE* FP;
unsigned char header[264];
int width, height, numslices, darkval, magic;
long numelems,i,j;
unsigned char *p, *strip;
int x,y;


	FP = fopen (filename, "rb");
	if (!FP) return -1;

	numelems = fread (header, sizeof (char), 264, FP);	/* read in header info */
	if (numelems != 264)
	{
		dp (0, "Error: WIM header: Read %d elements, should be 264\n", numelems);
		return -1;
	}
	magic = header[252]+ ((header[253])<<8);
	darkval = header[258] + ((header[259])<<8);
	height = header[260] + ((header[261])<<8);
	width = header[262] + ((header[263])<<8);
	numslices = 2;
	for (i=0; i<264; i++)
	{
		dp (2,"%5d   %3d  \n", i,header[i]);
	}
	
	dp (1, "WIM file header: Magic %d. Size %dx%d, dark value %d\n",magic, width,height,darkval);
	if (magic != 524) dp (0, "Warning: Possibly not a WIM file\n");
	
	/* We found the image size, now read strip by strip and sort into img->data */
	
	img->xmax = width;
	img->ymax = height; 
	img->zmax = numslices;
	img->maxval.imax=255;
	img->minval.imin=0;
	img->imgtype=B_CHAR;
	x = width;
	y = height;
	
	img->data = calloc (img->xmax*img->ymax*img->zmax,1);
	if (!img->data) 
	{
		fclose (FP);
		fatal_memerr();
	}
	p = img->data;

	strip = calloc (2*x+256, sizeof (unsigned char));
	if (!strip) fatal_memerr();

	
	/* Each strip contains pairs of bytes:
	   first byte, later slice 0, is high energy;
	   second byte, later lice 1, is low energy.
	*/
	
	for (i=0; i<y; i++)
	{
		numelems = fread (strip, sizeof (unsigned char), 2*x, FP);
		for (j=0; j<x; j++)
		{
			p[i*x+j] = strip[2*j];
			p[i*x+j+x*y] = strip[2*j+1];
		}
	}

	fclose (FP);
	free (strip);
	return 0;
	
}


int raw_write (image_type* img, const char* filename, int st, int en)
{
int e, errcode;
long numelems,l;
char* p;
FILE* FP;

	errcode=0;
	e = DATALEN (img->imgtype);
	l = e * img->xmax*img->ymax;
	p = (img->data) + l*st;
	
	FP = fopen (filename, "wb");
	if (!FP) return -1;
	
	numelems = fwrite (p, sizeof (char), l*(en-st+1), FP);
	if (numelems!=l*(en-st+1)) 
	{
		dp (1, "Rawrite warning: Fewer elements written than file size\n");
		errcode=-1;
	}
	fclose (FP);
	return errcode;
}



int DX_write (image_type* img, const char* filename, int st, int en, float zscale)
{
int e, errcode;
long numelems,l,zz;
char* p;
FILE* FP;
char header[32768];
char s[1024];
char* imgtypes[4] = { "byte","short","rgb","float" };

	errcode=0;
	e = DATALEN (img->imgtype);
	l = e * img->xmax*img->ymax;
	p = (img->data) + l*st;
	zz = en-st+1;
	
	FP = fopen (filename, "wb");
	if (!FP) return -1;
	
	if ((st==en) || (img->zmax==1))
	{
		/* Generate the DX header - 2D version */
	
		sprintf (header, "object 1 class array type %s rank 0 items %li lsb ieee data 0\n",
				imgtypes[img->imgtype],
				img->xmax*img->ymax );
		sprintf (s, "attribute \"dep\" string \"positions\"\n#\n");
		strcat (header, s);
		
		sprintf (s,"object 2 class gridpositions counts %li %li\n",img->ymax,img->xmax);
		strcat (header,s);
		/*sprintf (s, " origin 0 0\n delta 0 -1\n delta 1 0\n");*/
		sprintf (s, " origin %li %li\n delta 0 -1\n delta 1 0\n",-img->ymax/2,-img->xmax/2);
		strcat (header,s);
		sprintf (s, "attribute \"dep\" string \"positions\"\n#\n");
		strcat (header,s);

		sprintf (s, "object 3 class gridconnections counts %li %li\n",img->ymax,img->xmax);
		strcat (header,s);
		sprintf (s, "attribute \"element type\" string \"quads\"\n");
		strcat (header,s);
		sprintf (s, "attribute \"ref\" string \"positions\"\n#\n");
		strcat (header,s);
		
		sprintf (s, "object \"2D_image\" class field\n");
		strcat (header,s);
		sprintf (s, "component \"data\" value 1\n");
		strcat (header,s);
		sprintf (s, "component \"positions\" value 2\n");
		strcat (header,s);
		sprintf (s, "component \"connections\" value 3\n");
		strcat (header,s);
		sprintf (s, "attribute \"name\" string \"2D_image\"\n");
		strcat (header,s);
		sprintf (s, "#\nend\n");
		strcat (header,s);
	}	
	else
	{
		/* Generate the DX header - 3D version */
		/* TODO: Swaps X and Z coordinates */
	
		sprintf (header, "object 1 class array type %s rank 0 items %li lsb ieee data 0\n",
				imgtypes[img->imgtype],
				img->xmax*img->ymax*zz );
		sprintf (s, "attribute \"dep\" string \"positions\"\n#\n");
		strcat (header, s);
		
		sprintf (s,"object 2 class gridpositions counts %li %li %li\n",zz,img->ymax,img->xmax);
		strcat (header,s);
		/*sprintf (s, " origin 0.00 0.00 0.00\n delta %f 0.0 0.0\n delta 0.0 -1.0 0.0\n delta 0.0 0.0 1.0\n", zscale);*/
		sprintf (s, " origin %li %li %li\n delta %f 0 0\n delta 0 -1 0\n delta 0 0 1\n",-zz/2,-img->ymax/2,-img->xmax/2, zscale);
		strcat (header,s);
		sprintf (s, "attribute \"dep\" string \"positions\"\n#\n");
		strcat (header,s);

		sprintf (s, "object 3 class gridconnections counts %li %li %li\n",zz,img->ymax,img->xmax);
		strcat (header,s);
		sprintf (s, "attribute \"element type\" string \"cubes\"\n");
		strcat (header,s);
		sprintf (s, "attribute \"ref\" string \"positions\"\n#\n");
		strcat (header,s);
		
		sprintf (s, "object \"3D_image\" class field\n");
		strcat (header,s);
		sprintf (s, "component \"data\" value 1\n");
		strcat (header,s);
		sprintf (s, "component \"positions\" value 2\n");
		strcat (header,s);
		sprintf (s, "component \"connections\" value 3\n");
		strcat (header,s);
		sprintf (s, "attribute \"name\" string \"3D_image\"\n");
		strcat (header,s);
		sprintf (s, "#\nend\n");
		strcat (header,s);
	
	
	}
	
		
	/* Write the DX header */
	
	numelems = fwrite (header, sizeof (char), strlen(header), FP);
	
	/* Write raw data, same as in raw_write () */
	
	numelems = fwrite (p, sizeof (char), l*(en-st+1), FP);
	if (numelems!=l*(en-st+1)) 
	{
		dp (1, "DXwrite warning: Fewer elements written than file size\n");
		errcode=-1;
	}
	fclose (FP);
	return errcode;
}



/************************************************************************************

	Portable graphics formats, PGM and the likes.
	Note that netpbm files are single-slice only and we therefore
	save only the slice at st. The caller should make sure that the netpbm
	option is disabled for stacks.


************************************************************************************/



int pnm_write (image_type* img, const char* filename, int st, int en)
{
int e, errcode;
long numelems,l;
char* p;
FILE* FP;
char hdr[256];

	if ((img->imgtype==B_FLOAT) || (img->imgtype==B_SHORT)) return -1;

	if (img->imgtype==B_RGB)
		sprintf (hdr, "P6 %li %li %d\n", img->xmax, img->ymax, img->maxval.imax);
	else
		sprintf (hdr, "P5 %li %li %d\n", img->xmax, img->ymax, img->maxval.imax);

	errcode=0;
	e = DATALEN (img->imgtype);
	l = e * img->xmax*img->ymax;
	p = (img->data) + l*st;
	
	FP = fopen (filename, "wb");
	if (!FP) return -1;
	
	numelems = fwrite (hdr, sizeof (char), strlen(hdr), FP);
	
	numelems = fwrite (p, sizeof (char), l, FP);
	if (numelems!=l) 
	{
		dp (1, "Pnm write warning: Fewer elements written than file size\n");
		errcode=-1;
	}
	fclose (FP);
	return errcode;
}



int pnm_read (image_type* img, const char* filename)
{
int x,y, wspcnt,e;
FILE *FP;
int numelems, errcode;
char buf;
char hdr[256];
char dummy[256];
int xsize, ysize, maxgray;

	FP = fopen (filename, "rb");
	if (!FP) return -1;

	/* The portable netmap format has four white spaces. The header ends after the fourth.
	   The fourth whitespace must be a single whitespace. */

	x=0; y=0; wspcnt=0;	errcode=0;
	while ((wspcnt<4) && (x<250))
	{
		fread (&buf, 1, 1, FP);		/* One byte form the header */
		hdr[x] = buf;
		if ((buf==0x20) || (buf==0x09) || (buf==0x0D) || (buf==0x0A))	/* A whitespace */
		{
			if (y==0) wspcnt++;			/* If a new sequence of white spaces, increment white space count */
			y=1;						/* Flag ongoing white space */
		}
		else
			y=0;						/* End of white space */
		x++;
	}
	hdr[x]=0;							/* Terminate string */

	if (wspcnt<4) errcode= -1;			/* No four white spaces found in reasonable # of bytes */

	dp (2, "PNM: Read %d bytes of header information with a total of %d white spaces.\n",x, wspcnt);
	dp (2, "PNM: FYI - the header is %s\n",hdr);
	
	/* With the given information, allocate image memory & set up image */

	if (hdr[1] == 0x35)					/* PBM format */
		e=1;
	else if (hdr[1] == 0x36)				/* PPM format */
		e=3;
	else
		errcode=-1;

	if (errcode<0) dp (1, "PNM: Header error. Read failed.\n");
	if (errcode<0) return errcode;			/* Exit here if anything went wrong */

	sscanf (hdr,"%s %d %d %d\n", dummy, &xsize, &ysize, &maxgray);
	
	if (e==3)
	{
		if (allocate_image (img, B_RGB, xsize, ysize, 1)==-1) fatal_memerr();
		dp (2, "PNM: Allocated RGB image of %d x %d pixels\n",xsize, ysize);
	}
	else
	{
		if (allocate_image (img, B_CHAR, xsize, ysize, 1)==-1) fatal_memerr();
		dp (2, "PNM: Allocated UCHAR image of %d x %d pixels\n",xsize, ysize);
	}

	numelems = fread (img->data, sizeof(char), e*xsize*ysize, FP);
	fclose (FP);
	if (numelems!=e*xsize*ysize) 
	{
		dp (1, "Pnm read warning: Fewer elements obtained than file size indicates\n");
		errcode=-1;
	}

	return errcode;
}


/************************************************************************************

	ASCII write
	Since we can import ASCII, let's allow writing of ASCII data as well.
	ASCII files are huge, and we'll allow only one slice. Like pbm,
	the caller should make sure that the ASCII option is disabled for stacks.
	Note that we write float images with two decimals. If this is insufficient,
	multiply with some constant before saving.
	Note that RGB is output as R-R-R-...G-G-G-...B-B-B-B-...


************************************************************************************/



int ascii_write (image_type* img, const char* filename, int st, int en)
{
int e, errcode;
int x,y,z,cnt;
FILE* FP;


	errcode=0;
	
	FP = fopen (filename, "wt");
	if (!FP) return -1;
	
	if (img->imgtype==B_RGB)
	{
		for (z=st; z<=en; z++)
		{
			for (e=0; e<3; e++)
			{
				for (y=0; y<img->ymax; y++)
				{
					for (x=0; x<img->xmax; x++)
					{
						if (e==0) fprintf (FP, "%3d  ",((rgbtriplet*)img->data)[x+img->xmax*(y+img->ymax*z)].red);
						else if (e==1) fprintf (FP, "%3d  ",((rgbtriplet*)img->data)[x+img->xmax*(y+img->ymax*z)].green);
						else fprintf (FP, "%3d  ",((rgbtriplet*)img->data)[x+img->xmax*(y+img->ymax*z)].blue);
					}
					fprintf (FP, "\n"); 
				}
			}
		}
	}
	else if (img->imgtype==B_FLOAT)
	{
		for (z=st; z<=en; z++)
		{
			for (y=0; y<img->ymax; y++)
			{
				for (x=0; x<img->xmax; x++)
				{
					fprintf (FP, "%7.2f  ",((float*)img->data)[x+img->xmax*(y+img->ymax*z)]);
					cnt++;
				}
				fprintf (FP, "\n"); 
			}
		}
	}
	else if (img->imgtype==B_SHORT)
	{
		for (z=st; z<=en; z++)
		{
			for (y=0; y<img->ymax; y++)
			{
				for (x=0; x<img->xmax; x++)
				{
					fprintf (FP, "%5d  ",((short*)img->data)[x+img->xmax*(y+img->ymax*z)]);
				}
				fprintf (FP, "\n"); 
			}
		}
	}
	else if (img->imgtype==B_CHAR)
	{
		for (z=st; z<=en; z++)
		{
			for (y=0; y<img->ymax; y++)
			{
				for (x=0; x<img->xmax; x++)
				{
					fprintf (FP, "%3d  ",((unsigned char*)img->data)[x+img->xmax*(y+img->ymax*z)]);
				}
				fprintf (FP, "\n");
			}
		}
	}

	fclose (FP);
	return errcode;
}




/************************************************************************************

	In addition to is_tiff and is_vsrt, this function get_filetype tries to
	obtain the type of file by its header info. It encompasses is_tiff and
	is_vsrt.


************************************************************************************/

char get_filetype (const char* filename)
{
FILE* FP;
char hdr[256];
int numelems;
char ftype;

	FP = fopen (filename, "rb");
	if (!FP) return 0;
	
	numelems = fread (&hdr, sizeof (char), sizeof (hdr)-1, FP);
	fclose (FP);
	if (numelems < 128) return -1;		/* If this file is *that* short, it's not likely an image file */
	
	if (!strncasecmp (hdr, "VSRT", 4)) 
		ftype= FTYPE_VSRT;
	else if (!strncasecmp (hdr, "VFLT", 4)) 
		ftype= FTYPE_VFLT;
	else if (!strncmp (hdr, "II", 2)) 
		ftype= FTYPE_TIFFI; 			/* Little-endian, (II for Intel) */
	else if (!strncmp (hdr, "MM", 2)) 
		ftype= FTYPE_TIFFM;				/* Big-endian (MM for Motorola) */
	else if (!strncmp (hdr+0x80, "DICM", 4)) 
		ftype= FTYPE_DCM;				/* Dicom file has code at offset 128 */
	else if (!strncmp (hdr, "P5", 2)) 
		ftype= FTYPE_PGM;				/* PGM - portable graymap. One-byte binary grayscale */
	else if (!strncmp (hdr, "P6", 2)) 
		ftype= FTYPE_PPM;				/* PPM - portable bytemap. 24-bit RGB */
	else
		ftype=0;

	dp (1, "Probed file type. Type is %d ",ftype);
	if (ftype < NFILETYPES) dp (1, "(%s)", filetype_names[(unsigned char)ftype]);
	dp (1, "\n");


	return ftype;
	
}







/************************************************************************************

	And finally, although this is also not directly related to tiff,
	the save file dialog box and handler. I was just not willing
	to start a new file simply for this dialog.
	
************************************************************************************/

int save_type, compression_mode;
GtkAdjustment *savedlg_startslice_adj;
GtkAdjustment *savedlg_endslice_adj;
GtkAdjustment *savedlg_zscale_adj;
GtkWidget *savedlg_filename;
char savefilename[2048];
image_type* saveimg;
GtkWidget *savedlg_label2a, *savedlg_label2b;
GtkWidget *savedlg_zscale;
GtkWidget *savedlg_optionmenu2;



void on_savedlg_browse_clicked (GtkButton *button, gpointer user_data)
{
gchar* selectedfile; 

	ChooseFile ("Filename to save", window1, 1, &selectedfile);
	if (selectedfile)
	{
		strcpy (savefilename, selectedfile);
		gtk_entry_set_text (GTK_ENTRY(savedlg_filename), savefilename); 
	}
}


void on_savedlg_ok_clicked (GtkButton *button, gpointer user_data)
{
gchar* savefile;
int err;
int stslice, enslice;
float zscale;

	savefile = gtk_entry_get_text (GTK_ENTRY(savedlg_filename));
	err=0;		/* prevent a warning */
	stslice = (int) (0.5+savedlg_startslice_adj->value);
	savedlg_startslice_adj->value = stslice;
	enslice = (int) (0.5+savedlg_endslice_adj->value);
	if (enslice<stslice) enslice = stslice;
	savedlg_endslice_adj->value = enslice;
	zscale = savedlg_zscale_adj->value;
	
	if (save_type == 0)		/* TIFF */
	{
		dp (1, "Saving %s as TIFF, compression %d\n",savefile,compression_mode);
		err=tiffwrite (savefile, saveimg, compression_mode, "Made with Crystal Image 0.85 (C) M. Haidekker", stslice, enslice);
	}
	else if (save_type==1)		/* RAW */
	{
		dp (1, "Saving %s as RAW\n",savefile);
		err=raw_write (saveimg, savefile, stslice, enslice);
	}
	else if (save_type==2)		/* VSRT */
	{
		dp (1, "Saving %s as VSRT\n",savefile);
		err=vsrtwrite (saveimg, savefile, stslice, enslice);
	}
	else if (save_type==3)		/* OpenDX */
	{
		dp (1, "Saving %s as DX native\n",savefile);
		err=DX_write (saveimg, savefile, stslice, enslice, zscale);
	}
	else if (save_type==4)		/* PGM and PPM */
	{
		dp (1, "Saving %s as Portable Graymap PGM\n",savefile);
		if (stslice != enslice)
			ShowMessage ("Error: PGM cannot save stacks");
		else
			err=pnm_write (saveimg, savefile, stslice, enslice);
	}
	else if (save_type==5)		/* RAW ASCII */
	{
		dp (1, "Saving %s as raw ASCII file\n",savefile);
		err=ascii_write (saveimg, savefile, stslice, enslice);
	}
	
	
	if (err==0)
		statusbar_display ("File has been saved");
	else 
		statusbar_display ("File save possibly failed");
	set_display_title (savefile);

	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_savedlg_cancel_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}

void savedlg_set_type_menu (GtkWidget *widget, gpointer user_data)
{
	save_type = *(int*) user_data;
	if (save_type == 0)
	{
		gtk_widget_show (savedlg_label2a);
		gtk_widget_show (savedlg_optionmenu2);
		gtk_widget_hide (savedlg_label2b);
		gtk_widget_hide (savedlg_zscale);
	}
	else if (save_type == 3)
	{
		gtk_widget_hide (savedlg_label2a);
		gtk_widget_hide (savedlg_optionmenu2);
		gtk_widget_show (savedlg_label2b);
		gtk_widget_show (savedlg_zscale);
	}
	else
	{
		gtk_widget_hide (savedlg_label2a);
		gtk_widget_hide (savedlg_optionmenu2);
		gtk_widget_hide (savedlg_label2b);
		gtk_widget_hide (savedlg_zscale);
	}
	
}

void savedlg_set_compression_menu (GtkWidget *widget, gpointer user_data)
{
	compression_mode = *(int*) user_data;
}



void create_savedlg_wnd (image_type* img)
{
  GtkWidget *savedlg_wnd;
  GtkWidget *savedlg_table1;
  GtkWidget *savedlg_label1;
  GtkWidget *savedlg_optionmenu1;
  GtkWidget *savedlg_optionmenu1_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *savedlg_browse;
  GtkWidget *savedlg_optionmenu2_menu;
  GtkWidget *savedlg_label3;
  GtkWidget *savedlg_label4;
  GtkWidget *savedlg_startslice;
  GtkWidget *savedlg_endslice;
  GtkWidget *savedlg_okbutton;
  GtkWidget *savedlg_cancel;

	savedlg_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (savedlg_wnd), "savedlg_wnd", savedlg_wnd);
	gtk_window_set_title (GTK_WINDOW (savedlg_wnd), "Save file as");
	gtk_window_set_modal (GTK_WINDOW (savedlg_wnd), TRUE);

	savedlg_table1 = gtk_table_new (7, 2, FALSE);
	gtk_widget_ref (savedlg_table1);
	gtk_object_set_data_full (GTK_OBJECT (savedlg_wnd), "savedlg_table1", savedlg_table1,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (savedlg_table1);
	gtk_container_add (GTK_CONTAINER (savedlg_wnd), savedlg_table1);

	savedlg_label1 = MakeLabel (savedlg_wnd,savedlg_table1,"File Format","savedlg_label1",0,0);
	savedlg_label2b = MakeLabel (savedlg_wnd,savedlg_table1,"Z anisotropy","savedlg_label2b",0,3);
	gtk_widget_hide (savedlg_label2b);
	savedlg_label2a = MakeLabel (savedlg_wnd,savedlg_table1,"TIFF compression","savedlg_label2a",0,2);
	savedlg_label3 = MakeLabel (savedlg_wnd,savedlg_table1,"Start Slice","savedlg_label3",0,4);
	savedlg_label4 = MakeLabel (savedlg_wnd,savedlg_table1,"End Slice","savedlg_label3",0,5);

	savedlg_browse = Button_in_table (savedlg_wnd,savedlg_table1,"Browse Filename","savedlg_browse",0,1,on_savedlg_browse_clicked);
	savedlg_okbutton = Button_in_table (savedlg_wnd,savedlg_table1,"Save","savedlg_okbutton",0,6,on_savedlg_ok_clicked);
	savedlg_cancel = Button_in_table (savedlg_wnd,savedlg_table1,"Cancel","savedlg_cancel",1,6,on_savedlg_cancel_clicked);


	savedlg_optionmenu1 = gtk_option_menu_new ();
	gtk_widget_ref (savedlg_optionmenu1);
	gtk_object_set_data_full (GTK_OBJECT (savedlg_wnd), "savedlg_optionmenu1", savedlg_optionmenu1,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (savedlg_optionmenu1);
	gtk_table_attach (GTK_TABLE (savedlg_table1), savedlg_optionmenu1, 1, 2, 0, 1,
                      (GtkAttachOptions) (0), (GtkAttachOptions) (0), 3, 3);
	gtk_container_set_border_width (GTK_CONTAINER (savedlg_optionmenu1), 6);

	savedlg_optionmenu1_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"TIFF",&menuopt0);
	glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"RAW",&menuopt1);
	glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"RAW ASCII",&menuopt5);
	if ((img->imgtype == B_SHORT) || (img->imgtype == B_FLOAT))
  	  glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"VSRT/VFLT",&menuopt2);
	if (img->imgtype != B_RGB)
  	  glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"DX",&menuopt3);
	if (img->imgtype == B_CHAR)
  	  glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"PGM",&menuopt4);
	else if (img->imgtype == B_RGB)
  	  glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"PPM",&menuopt4);

	gtk_option_menu_set_menu (GTK_OPTION_MENU (savedlg_optionmenu1), savedlg_optionmenu1_menu);

	savedlg_optionmenu2 = gtk_option_menu_new ();
	gtk_widget_ref (savedlg_optionmenu2);
	gtk_object_set_data_full (GTK_OBJECT (savedlg_wnd), "savedlg_optionmenu2", savedlg_optionmenu2,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (savedlg_optionmenu2);
	gtk_table_attach (GTK_TABLE (savedlg_table1), savedlg_optionmenu2, 1, 2, 2, 3,
                      (GtkAttachOptions) (0),
                      (GtkAttachOptions) (0), 0, 0);
	gtk_container_set_border_width (GTK_CONTAINER (savedlg_optionmenu2), 6);

	savedlg_optionmenu2_menu = gtk_menu_new ();

	/* The choice of TIFF compression schemes actually depends on the image type we have */

	switch (img->imgtype)
	{
		case B_CHAR:
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"NONE",&menuopt0);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"LZW",&menuopt1);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"Packbits",&menuopt2);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"CCITT FAX 3",&menuopt3);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"CCITT FAX 4",&menuopt4);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"JPEG",&menuopt5);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"CCITT RLE",&menuopt6);
			break;
  		case B_SHORT:
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"NONE",&menuopt0);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"LZW",&menuopt1);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"Packbits",&menuopt2);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"JPEG",&menuopt5);
			break;
  		case B_FLOAT:
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"NONE",&menuopt0);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"LZW",&menuopt1);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"Packbits",&menuopt2);
			break;
  		case B_RGB:
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"NONE",&menuopt0);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"LZW",&menuopt1);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"Packbits",&menuopt2);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"JPEG",&menuopt5);
			break;
	}		

	gtk_option_menu_set_menu (GTK_OPTION_MENU (savedlg_optionmenu2), savedlg_optionmenu2_menu);

	savedlg_filename = gtk_entry_new ();
	gtk_widget_ref (savedlg_filename);
	gtk_object_set_data_full (GTK_OBJECT (savedlg_wnd), "savedlg_filename", savedlg_filename,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (savedlg_filename);
	gtk_table_attach (GTK_TABLE (savedlg_table1), savedlg_filename, 1, 2, 1, 2,
                      (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                      (GtkAttachOptions) (0), 6, 6);

	savedlg_zscale_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 0, 25, 0.1, 1, 0));
	savedlg_zscale = MakeHScale (savedlg_zscale_adj, savedlg_wnd, savedlg_table1,
  									  "savedlg_zscale", 1,3);
	gtk_widget_hide (savedlg_zscale);

	savedlg_startslice_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, 0, img->zmax-1, 1, 10, 0));
	savedlg_startslice = MakeHScale (savedlg_startslice_adj, savedlg_wnd, savedlg_table1,
  									  "savedlg_startslice", 1,4);

	savedlg_endslice_adj = GTK_ADJUSTMENT(gtk_adjustment_new (img->zmax-1, 0, img->zmax-1, 1, 10, 0));
	savedlg_endslice = MakeHScale (savedlg_endslice_adj, savedlg_wnd, savedlg_table1,
  									  "savedlg_startslice", 1,5);

	gtk_scale_set_digits (GTK_SCALE (savedlg_startslice), 0);
	gtk_scale_set_digits (GTK_SCALE (savedlg_endslice), 0);

	gtk_signal_connect (GTK_OBJECT (savedlg_wnd), "destroy",
                    	GTK_SIGNAL_FUNC (on_savedlg_cancel_clicked),savedlg_wnd);

	if (mainimg_name[0]) gtk_entry_set_text (GTK_ENTRY(savedlg_filename), mainimg_name);

	gtk_widget_show ( savedlg_wnd);
}


void save_main (image_type* img)
{

	save_type=0;
	compression_mode=0;
	saveimg = img;
	
	create_savedlg_wnd (img);
	
}


/*******************************************************************************/



